from __future__ import annotations

import json

import anyio
from mcp.types import CompletionArgument, ResourceTemplateReference
import pytest

import refua_mcp.server as server


def test_capabilities_resource_reports_core_runtime_flags() -> None:
    payload = json.loads(server.refua_capabilities())
    assert payload["mcp_spec_revision"] == server.MCP_SPEC_REVISION
    assert payload["mcp_latest_protocol_version"] == server.LATEST_PROTOCOL_VERSION
    assert "features" in payload
    assert "runtime" in payload
    assert payload["runtime"]["task_timeout_seconds"] >= 0
    assert payload["runtime"]["queue_timeout_seconds"] >= 0
    assert (
        payload["features"]["clinical_simulator_available"]
        == server._CLINICAL_AVAILABLE
    )
    assert payload["features"]["data_available"] == server._DATA_AVAILABLE
    assert payload["features"]["preclinical_available"] == server._PRECLINICAL_AVAILABLE


def test_protein_property_resources_are_exposed() -> None:
    index = json.loads(server.refua_protein_property_index())
    assert "property_names" in index
    assert "property_groups" in index
    assert index["count"]["properties"] == len(index["property_names"])


def test_recipe_index_conditionally_includes_clinical_recipe() -> None:
    payload = json.loads(server.refua_recipe_index())
    names = set(payload["recipe_names"])
    if server._CLINICAL_AVAILABLE:
        assert "clinical_simulation" in names
    else:
        assert "clinical_simulation" not in names
    if server._DATA_AVAILABLE:
        assert "data_materialize" in names
        assert "data_query" in names
    else:
        assert "data_materialize" not in names
        assert "data_query" not in names
    if server._PRECLINICAL_AVAILABLE:
        assert "preclinical_plan" in names
        assert "preclinical_workup" in names
        assert "preclinical_cmc_plan" in names
        assert "preclinical_cmc_release" in names
    else:
        assert "preclinical_plan" not in names
        assert "preclinical_workup" not in names
        assert "preclinical_cmc_plan" not in names
        assert "preclinical_cmc_release" not in names


def test_recipe_completion_suggests_known_recipes() -> None:
    async def _run() -> None:
        completion = await server.refua_completion(
            ResourceTemplateReference(
                type="ref/resource",
                uri="refua://recipes/{recipe_name}",
            ),
            CompletionArgument(name="recipe_name", value="pro"),
            None,
        )
        assert completion is not None
        assert "protein_properties" in completion.values

    anyio.run(_run)


def test_protein_property_completion_suggests_groups() -> None:
    async def _run() -> None:
        completion = await server.refua_completion(
            ResourceTemplateReference(
                type="ref/resource",
                uri="refua://protein-properties/group/{group_name}",
            ),
            CompletionArgument(name="group_name", value="ba"),
            None,
        )
        assert completion is not None
        assert any(value.startswith("ba") for value in completion.values)

    anyio.run(_run)


def test_runtime_config_parses_secure_streamable_http(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("REFUA_MCP_TRANSPORT", "streamable-http")
    monkeypatch.setenv("REFUA_MCP_HOST", "127.0.0.1")
    monkeypatch.setenv("REFUA_MCP_PORT", "9100")
    monkeypatch.setenv("REFUA_MCP_AUTH_TOKENS", "token-a,token-b")
    monkeypatch.setenv("REFUA_MCP_ALLOWED_HOSTS", "127.0.0.1:9100")
    monkeypatch.setenv("REFUA_MCP_ALLOWED_ORIGINS", "http://127.0.0.1:9100")
    config = server._build_runtime_server_config()

    assert config.transport == "streamable-http"
    assert config.token_count == 2
    assert config.enable_dns_rebinding_protection is True
    assert config.allowed_hosts == ("127.0.0.1:9100",)
    assert config.allowed_origins == ("http://127.0.0.1:9100",)



from __future__ import annotations

import json
import os
import sys
from pathlib import Path

import anyio
import mcp.types as types
from mcp.client.session import ClientSession
from mcp.client.stdio import StdioServerParameters, stdio_client

import refua_mcp.server as server

ROOT = Path(__file__).resolve().parents[1]


def _server_python() -> str:
    override = os.environ.get("REFUA_MCP_TEST_SERVER_PYTHON")
    if override:
        return override
    return sys.executable


async def _with_session() -> tuple[ClientSession, object]:
    params = StdioServerParameters(
        command=_server_python(),
        args=["-m", "refua_mcp.server"],
        cwd=ROOT,
        env={
            "PYTHONPATH": str(ROOT / "src"),
            "REFUA_MCP_TRANSPORT": "stdio",
        },
    )
    client_cm = stdio_client(params)
    read_stream, write_stream = await client_cm.__aenter__()
    session_cm = ClientSession(read_stream, write_stream)
    session = await session_cm.__aenter__()

    class _Closer:
        async def aclose(self) -> None:
            await session_cm.__aexit__(None, None, None)
            await client_cm.__aexit__(None, None, None)

    return session, _Closer()


def test_sdk_acceptance_initialize_and_discovery() -> None:
    async def _run() -> None:
        session, closer = await _with_session()
        try:
            init = await session.initialize()
            assert init.protocolVersion == server.MCP_SPEC_REVISION

            tools = await session.list_tools()
            tool_names = {tool.name for tool in tools.tools}
            assert "refua_protein_properties" in tool_names
            assert "refua_fold" in tool_names
            if server._CLINICAL_AVAILABLE:
                assert "refua_clinical_simulator" in tool_names
            else:
                assert "refua_clinical_simulator" not in tool_names
            if server._PRECLINICAL_AVAILABLE:
                assert "refua_preclinical_workup" in tool_names
                assert "refua_preclinical_cmc_plan" in tool_names
            else:
                assert "refua_preclinical_workup" not in tool_names
                assert "refua_preclinical_cmc_plan" not in tool_names

            templates = await session.list_resource_templates()
            template_uris = {
                template.uriTemplate for template in templates.resourceTemplates
            }
            assert "refua://recipes/{recipe_name}" in template_uris
            assert "refua://protein-properties/group/{group_name}" in template_uris
            assert (
                "refua://protein-properties/property/{property_name}" in template_uris
            )

            capabilities = await session.read_resource("refua://capabilities")
            text = capabilities.contents[0].text
            payload = json.loads(text)
            assert payload["mcp_spec_revision"] == server.MCP_SPEC_REVISION
        finally:
            await closer.aclose()

    anyio.run(_run)


def test_sdk_acceptance_completion_and_tool_call() -> None:
    async def _run() -> None:
        session, closer = await _with_session()
        try:
            await session.initialize()

            recipe_completion = await session.complete(
                types.ResourceTemplateReference(
                    type="ref/resource",
                    uri="refua://recipes/{recipe_name}",
                ),
                {"name": "recipe_name", "value": "pro"},
            )
            assert "protein_properties" in recipe_completion.completion.values

            result = await session.call_tool(
                "refua_protein_properties",
                {
                    "sequence": "MKTAYIAKQRQISFVKSHFSRQLEERLGLIEVQ",
                    "groups": ["basic"],
                },
            )
            assert result.isError is False
            assert isinstance(result.structuredContent, dict)
            assert "values" in result.structuredContent
        finally:
            await closer.aclose()

    anyio.run(_run)


def test_sdk_acceptance_task_augmented_validate_spec() -> None:
    async def _run() -> None:
        session, closer = await _with_session()
        try:
            await session.initialize()
            task = await session.experimental.call_tool_as_task(
                "refua_validate_spec",
                {
                    "entities": [
                        {
                            "type": "protein",
                            "id": "A",
                            "sequence": "MKTAYIAK",
                        }
                    ]
                },
                ttl=120000,
            )
            task_id = task.task.taskId

            terminal = None
            for _ in range(200):
                status = await session.experimental.get_task(task_id)
                if status.status in {"completed", "failed", "cancelled"}:
                    terminal = status
                    break
                await anyio.sleep(0.01)

            assert terminal is not None
            assert terminal.status == "completed"

            final = await session.experimental.get_task_result(
                task_id,
                types.CallToolResult,
            )
            assert final.isError is False
            assert isinstance(final.structuredContent, dict)
            assert final.structuredContent.get("valid") is True
        finally:
            await closer.aclose()

    anyio.run(_run)



from __future__ import annotations

import sys
import types

import pytest

import refua_mcp.server as server


def _install_fake_refua(monkeypatch: pytest.MonkeyPatch) -> None:
    module = types.ModuleType("refua")

    class FakeProteinProperties:
        def __init__(self, sequence: str, *, lazy: bool = True, sanitize: bool = True):
            del lazy
            self.sequence = (
                "".join(sequence.split()).upper() if sanitize else str(sequence)
            )

        @classmethod
        def from_sequence(
            cls,
            sequence: str,
            *,
            lazy: bool = True,
            sanitize: bool = True,
        ) -> "FakeProteinProperties":
            return cls(sequence, lazy=lazy, sanitize=sanitize)

        def get(self, name: str):
            key = str(name).strip().lower()
            if key == "pi":
                key = "isoelectric_point"
            values = self.to_dict()
            if key not in values:
                raise KeyError(name)
            return values[key]

        def to_dict(self, *, groups=None):
            all_values = {
                "length": len(self.sequence),
                "isoelectric_point": 7.2,
                "gravy": -0.21,
                "net_charge_ph_7_4": 0.5,
            }
            if groups is None:
                return dict(all_values)
            group_members = {
                "basic": {"length", "isoelectric_point", "gravy"},
                "charge": {"isoelectric_point", "net_charge_ph_7_4"},
            }
            selected: set[str] = set()
            for group in groups:
                selected.update(group_members.get(str(group).lower(), set()))
            return {name: all_values[name] for name in all_values if name in selected}

    module.ProteinProperties = FakeProteinProperties
    module.available_protein_properties = lambda: (
        "length",
        "isoelectric_point",
        "gravy",
        "net_charge_ph_7_4",
    )
    module.available_protein_property_groups = lambda: ("basic", "charge")
    monkeypatch.setitem(sys.modules, "refua", module)


def test_refua_protein_properties_tool_is_registered() -> None:
    tool = next(
        info
        for info in server.mcp._tool_manager.list_tools()
        if info.name == "refua_protein_properties"
    )
    assert "sequence" in tool.parameters["properties"]
    assert tool.parameters["properties"]["include_catalog"]["type"] == "boolean"


def test_refua_protein_properties_supports_property_filter_and_catalog(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _install_fake_refua(monkeypatch)
    result = server.refua_protein_properties(
        sequence="mk ta",
        properties='["length", "pi"]',
        include_catalog=True,
    )

    assert result.sequence == "mk ta"
    assert result.normalized_sequence == "MKTA"
    assert result.selected_properties == ["length", "pi"]
    assert result.values == {"length": 4, "pi": 7.2}
    assert result.available_properties == [
        "gravy",
        "isoelectric_point",
        "length",
        "net_charge_ph_7_4",
    ]
    assert result.available_property_groups == ["basic", "charge"]


def test_refua_protein_properties_supports_group_filter_from_string(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _install_fake_refua(monkeypatch)
    result = server.refua_protein_properties(
        sequence="MKTAY",
        groups="basic,charge",
    )

    assert result.selected_groups == ["basic", "charge"]
    assert result.values["length"] == 5
    assert result.values["isoelectric_point"] == 7.2
    assert result.values["net_charge_ph_7_4"] == 0.5


def test_refua_protein_properties_rejects_mixed_group_and_property_filters() -> None:
    with pytest.raises(ValueError, match="either properties or groups"):
        server.refua_protein_properties(
            sequence="MKTAY",
            properties=["length"],
            groups=["basic"],
        )



"""Refua MCP server package."""

from importlib.metadata import version as _distribution_version
from pathlib import Path
import tomllib

__all__ = ["__version__"]


def _read_version_from_pyproject() -> str | None:
    pyproject_path = Path(__file__).resolve().parents[2] / "pyproject.toml"
    if not pyproject_path.exists():
        return None
    data = tomllib.loads(pyproject_path.read_text(encoding="utf-8"))
    project = data.get("project", {})
    version = project.get("version")
    if not version:
        return None
    return str(version)


def _resolve_version() -> str:
    local_version = _read_version_from_pyproject()
    if local_version is not None:
        return local_version
    return _distribution_version("refua-mcp")


__version__ = _resolve_version()



from __future__ import annotations

import mcp.types as types

import refua_mcp.server as server


def test_protocol_target_revision_matches_sdk_latest() -> None:
    assert server.MCP_SPEC_REVISION == "2025-11-25"
    assert str(types.LATEST_PROTOCOL_VERSION) == server.MCP_SPEC_REVISION


def test_tasks_handlers_are_registered_for_experimental_support() -> None:
    lowlevel = server.mcp._mcp_server
    assert lowlevel.experimental.task_support is not None

    required_handlers = {
        types.GetTaskRequest,
        types.GetTaskPayloadRequest,
        types.ListTasksRequest,
        types.CancelTaskRequest,
    }
    assert required_handlers.issubset(set(lowlevel.request_handlers))



from __future__ import annotations

import json
from types import SimpleNamespace

import anyio
from mcp.types import CallToolResult

import refua_mcp.server as server


def test_clinical_tool_registration_matches_dependency() -> None:
    tool_names = {info.name for info in server.mcp._tool_manager.list_tools()}
    if server._CLINICAL_AVAILABLE:
        assert "refua_clinical_simulator" in tool_names
    else:
        assert "refua_clinical_simulator" not in tool_names


def test_data_tool_registration_matches_dependency() -> None:
    tool_names = {info.name for info in server.mcp._tool_manager.list_tools()}
    data_tools = {
        "refua_data_list",
        "refua_data_fetch",
        "refua_data_materialize",
        "refua_data_query",
    }
    if server._DATA_AVAILABLE:
        assert data_tools.issubset(tool_names)
    else:
        assert data_tools.isdisjoint(tool_names)


def test_preclinical_tool_registration_matches_dependency() -> None:
    tool_names = {info.name for info in server.mcp._tool_manager.list_tools()}
    preclinical_tools = {
        "refua_preclinical_templates",
        "refua_preclinical_plan",
        "refua_preclinical_schedule",
        "refua_preclinical_bioanalysis",
        "refua_preclinical_workup",
        "refua_preclinical_cmc_templates",
        "refua_preclinical_cmc_plan",
        "refua_preclinical_batch_record",
        "refua_preclinical_stability_plan",
        "refua_preclinical_stability_assess",
        "refua_preclinical_release_assess",
    }
    if server._PRECLINICAL_AVAILABLE:
        assert preclinical_tools.issubset(tool_names)
    else:
        assert preclinical_tools.isdisjoint(tool_names)


def test_refua_fold_schema_uses_discriminated_entities() -> None:
    tool = next(
        info
        for info in server.mcp._tool_manager.list_tools()
        if info.name == "refua_fold"
    )
    entities_schema = tool.parameters["properties"]["entities"]
    array_variant = entities_schema["anyOf"][0]
    entities_items = array_variant["items"]["anyOf"][0]
    assert "oneOf" in entities_items
    assert entities_items["discriminator"]["propertyName"] == "type"


def test_validate_spec_returns_execution_plan() -> None:
    result = server.refua_validate_spec(
        entities=[
            server.ProteinEntity(type="protein", id="A", sequence="MKTAYIAK"),
            server.LigandEntity(type="ligand", id="lig", smiles="CCO"),
        ],
        action="fold",
    )

    assert result.valid is True
    assert result.execution_plan.action == "fold"
    assert result.execution_plan.run_boltz is True
    assert result.execution_plan.run_boltzgen is False
    assert result.execution_plan.ligand_id_map == {"lig": "L1"}
    assert any("deep_validate=true" in warning for warning in result.warnings)


def test_recipe_template_returns_split_tool_examples() -> None:
    payload = json.loads(server.refua_recipe_template("antibody_design"))
    assert payload["tool"] == "refua_antibody_design"


def test_antibody_design_schema_exposes_output_format_compat_aliases() -> None:
    tool = next(
        info
        for info in server.mcp._tool_manager.list_tools()
        if info.name == "refua_antibody_design"
    )
    structure_enum = set(
        tool.parameters["properties"]["structure_output_format"]["anyOf"][0]["enum"]
    )
    feature_enum = set(
        tool.parameters["properties"]["feature_output_format"]["anyOf"][0]["enum"]
    )
    assert structure_enum == {"cif", "mmcif", "bcif"}
    assert feature_enum == {"torch", "npz", "json"}


def test_resolve_output_format_maps_mmcif_alias_to_cif() -> None:
    assert server._resolve_output_format(None, "mmcif") == "cif"


def test_json_feature_output_format_errors_only_for_file_writes() -> None:
    try:
        server._resolve_feature_output_format("features.pt", "json")
        raise AssertionError("Expected ValueError for json feature file output format.")
    except ValueError as exc:
        assert "inline MCP response summaries" in str(exc)

    result = server.refua_validate_spec(
        entities=[server.ProteinEntity(type="protein", id="A", sequence="MKTAYIAK")],
        feature_output_format="json",
    )
    assert any(
        "ignored unless feature_output_path" in warning for warning in result.warnings
    )


def test_validate_spec_normalizes_json_feature_file_output_request() -> None:
    result = server.refua_validate_spec(
        entities=[server.ProteinEntity(type="protein", id="A", sequence="MKTAYIAK")],
        feature_output_path="design_features.json",
        feature_output_format="json",
    )
    assert result.normalized_input["feature_output_format"] == "npz"
    assert result.normalized_input["feature_output_path"].endswith(
        "design_features.npz"
    )
    assert any("normalized to 'npz'" in warning for warning in result.warnings)


def test_validate_spec_normalizes_json_feature_path_without_format() -> None:
    result = server.refua_validate_spec(
        entities=[server.ProteinEntity(type="protein", id="A", sequence="MKTAYIAK")],
        feature_output_path="design_features.json",
    )
    assert result.normalized_input["feature_output_format"] == "npz"
    assert result.normalized_input["feature_output_path"].endswith(
        "design_features.npz"
    )
    assert any("ending in '.json'" in warning for warning in result.warnings)


def test_validate_spec_normalizes_mmcif_alias() -> None:
    result = server.refua_validate_spec(
        entities=[server.ProteinEntity(type="protein", id="A", sequence="MKTAYIAK")],
        structure_output_format="mmcif",
    )
    assert result.normalized_input["structure_output_format"] == "cif"
    assert any("normalized to 'cif'" in warning for warning in result.warnings)


def test_refua_fold_normalizes_output_requests_before_execution() -> None:
    captured: dict[str, object] = {}
    original = server._run_complex_operation

    def fake_run_complex_operation(**kwargs):
        captured.update(kwargs)
        return {
            "name": kwargs["name"],
            "backend": "test-backend",
            "chain_ids": [],
            "binder_sequences": {},
            "warnings": list(kwargs.get("output_warnings") or []),
        }

    server._run_complex_operation = fake_run_complex_operation
    try:
        result = server.refua_fold(
            name="production_design",
            entities=[{"type": "protein", "id": "A", "sequence": "MKTAYIAK"}],
            feature_output_path="design_features.json",
            feature_output_format="json",
            structure_output_format="mmcif",
            run_boltz=False,
            run_boltzgen=False,
        )
    finally:
        server._run_complex_operation = original

    assert captured["feature_output_format"] == "npz"
    assert str(captured["feature_output_path"]).endswith("design_features.npz")
    assert captured["structure_output_format"] == "cif"
    assert any("normalized to 'npz'" in warning for warning in result.warnings)
    assert any("normalized to 'cif'" in warning for warning in result.warnings)


def test_antibody_design_arg_model_accepts_json_string_compat_inputs() -> None:
    arg_model = server.mcp._tool_manager._tools[
        "refua_antibody_design"
    ].fn_metadata.arg_model
    parsed = arg_model(
        name="schema_probe",
        antibody='{"type":"antibody","ids":["H","L"],"heavy_cdr_lengths":[12,10,14],"light_cdr_lengths":[10,9,9]}',
        context_entities='[{"type":"protein","id":"A","sequence":"MKTAYIAK"}]',
        run_boltz="false",
        run_boltzgen="true",
        return_mmcif=False,
        async_mode=True,
    )
    assert parsed.antibody.startswith("{")
    assert parsed.context_entities.startswith("[")
    assert parsed.run_boltz is False
    assert parsed.run_boltzgen is True


def test_fold_arg_model_accepts_stringified_entity_entries() -> None:
    arg_model = server.mcp._tool_manager._tools["refua_fold"].fn_metadata.arg_model
    parsed = arg_model(
        entities=['{"type":"protein","id":"A","sequence":"MKTAYIAK"}'],
        run_boltz="true",
        run_boltzgen="false",
    )
    assert isinstance(parsed.entities, list)
    assert isinstance(parsed.entities[0], str)
    assert parsed.run_boltz is True
    assert parsed.run_boltzgen is False


def test_non_task_wrapper_returns_structured_content() -> None:
    async def _run() -> None:
        original_get_context = server.mcp.get_context
        try:
            server.mcp.get_context = lambda: SimpleNamespace(
                request_context=SimpleNamespace(experimental=None)
            )
            result = await server._call_tool_with_task_support(
                "refua_validate_spec",
                {
                    "entities": [
                        {"type": "protein", "id": "A", "sequence": "MKTAYIAK"},
                    ]
                },
            )
        finally:
            server.mcp.get_context = original_get_context

        assert isinstance(result, CallToolResult)
        assert isinstance(result.structuredContent, dict)
        assert result.structuredContent.get("valid") is True

    anyio.run(_run)


def test_non_task_wrapper_returns_standard_error_contract() -> None:
    async def _run() -> None:
        original_get_context = server.mcp.get_context
        try:
            server.mcp.get_context = lambda: SimpleNamespace(
                request_context=SimpleNamespace(experimental=None)
            )
            result = await server._call_tool_with_task_support(
                "refua_protein_properties",
                {"sequence": "   "},
            )
        finally:
            server.mcp.get_context = original_get_context

        assert isinstance(result, CallToolResult)
        assert result.isError is True
        assert isinstance(result.structuredContent, dict)
        error = result.structuredContent.get("error", {})
        assert error.get("code") == "invalid_input"
        assert "hint" in error

    anyio.run(_run)


def test_exploratory_name_guard_blocks_probe_patterns() -> None:
    for name in ("sanity_antibody", "schema_probe", "fold_probe2", "smoke_run"):
        try:
            server._enforce_non_exploratory_execution(
                name,
                allow_exploratory_run=False,
            )
            raise AssertionError("Expected exploratory-name guard to raise.")
        except ValueError as exc:
            assert "blocked by default" in str(exc)


def test_exploratory_name_guard_can_be_overridden() -> None:
    server._enforce_non_exploratory_execution(
        "schema_probe",
        allow_exploratory_run=True,
    )


def test_fold_schema_includes_exploratory_override_flag() -> None:
    tool = next(
        info
        for info in server.mcp._tool_manager.list_tools()
        if info.name == "refua_fold"
    )
    flag = tool.parameters["properties"]["allow_exploratory_run"]
    assert flag["type"] == "boolean"
    assert flag["default"] is False



from __future__ import annotations

import sys
import types

import refua_mcp.server as server


def _install_fake_refua(monkeypatch):
    module = types.ModuleType("refua")

    class FakeProtein:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    class FakeDNA:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    class FakeRNA:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    class FakeBinder:
        def __init__(
            self,
            spec=None,
            length=None,
            template_values=None,
            ids=None,
            binding_types=None,
            secondary_structure=None,
            cyclic=False,
        ):
            self.spec = spec
            self.length = length
            self.template_values = template_values
            self.ids = ids
            self.binding_types = binding_types
            self.secondary_structure = secondary_structure
            self.cyclic = cyclic

    class FakeComplex:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            self.added = []
            self.file_calls = []

        def add(self, *entities):
            self.added.extend(entities)
            return self

        def file(self, *args, **kwargs):
            self.file_calls.append((args, kwargs))
            return self

    class FakePair:
        def __init__(self, heavy, light):
            self.heavy = heavy
            self.light = light

    class FakeBinderDesigns:
        peptide_calls = []
        disulfide_calls = []
        antibody_calls = []

        @staticmethod
        def peptide(**kwargs):
            FakeBinderDesigns.peptide_calls.append(dict(kwargs))
            return FakeBinder(
                spec=str(kwargs.get("length", 12)),
                ids=kwargs.get("ids", "P"),
                binding_types=kwargs.get("binding_types"),
                secondary_structure=kwargs.get("secondary_structure"),
                cyclic=bool(kwargs.get("cyclic", False)),
            )

        @staticmethod
        def disulfide_peptide(**kwargs):
            FakeBinderDesigns.disulfide_calls.append(dict(kwargs))
            return FakeBinder(
                spec="10C6C3",
                ids=kwargs.get("ids", "P"),
                binding_types=kwargs.get("binding_types"),
                secondary_structure=kwargs.get("secondary_structure"),
                cyclic=bool(kwargs.get("cyclic", True)),
            )

        @staticmethod
        def antibody(**kwargs):
            FakeBinderDesigns.antibody_calls.append(dict(kwargs))
            heavy = FakeBinder(
                spec="heavy-default",
                ids=kwargs.get("heavy_id", "H"),
                binding_types=kwargs.get("heavy_binding_types"),
                secondary_structure=kwargs.get("heavy_secondary_structure"),
                cyclic=bool(kwargs.get("heavy_cyclic", False)),
            )
            light = FakeBinder(
                spec="light-default",
                ids=kwargs.get("light_id", "L"),
                binding_types=kwargs.get("light_binding_types"),
                secondary_structure=kwargs.get("light_secondary_structure"),
                cyclic=bool(kwargs.get("light_cyclic", False)),
            )
            return FakePair(heavy=heavy, light=light)

    class FakeSmallMolecule:
        @staticmethod
        def from_smiles(smiles):
            return f"smiles:{smiles}"

        @staticmethod
        def from_mol(mol, name):
            return f"mol:{name}"

    module.Protein = FakeProtein
    module.DNA = FakeDNA
    module.RNA = FakeRNA
    module.Binder = FakeBinder
    module.BinderDesigns = FakeBinderDesigns
    module.Complex = FakeComplex
    module.SmallMolecule = FakeSmallMolecule
    monkeypatch.setitem(sys.modules, "refua", module)
    return module


def test_binder_template_values_passthrough(monkeypatch):
    fake_refua = _install_fake_refua(monkeypatch)
    complex_spec, _, has_boltz, has_boltzgen = server._build_complex_from_spec(
        name="templated",
        base_dir=None,
        entities=[
            {
                "type": "binder",
                "id": "B",
                "spec": "{core}C{loop}",
                "template_values": {"core": 10, "loop": 6},
            }
        ],
        boltz_mol_dir=None,
    )

    assert has_boltz is False
    assert has_boltzgen is True
    assert len(complex_spec.added) == 1
    binder = complex_spec.added[0]
    assert isinstance(binder, fake_refua.Binder)
    assert binder.template_values == {"core": 10, "loop": 6}
    assert binder.ids == "B"


def test_peptide_disulfide_helper(monkeypatch):
    fake_refua = _install_fake_refua(monkeypatch)
    complex_spec, _, _, has_boltzgen = server._build_complex_from_spec(
        name="peptide",
        base_dir=None,
        entities=[
            {
                "type": "peptide",
                "id": "P",
                "segment_lengths": [8, 5, 4],
                "disulfide": True,
            }
        ],
        boltz_mol_dir=None,
    )

    assert has_boltzgen is True
    assert len(complex_spec.added) == 1
    assert len(fake_refua.BinderDesigns.disulfide_calls) == 1
    call = fake_refua.BinderDesigns.disulfide_calls[0]
    assert call["segment_lengths"] == (8, 5, 4)
    assert call["ids"] == "P"


def test_antibody_helper_with_overrides(monkeypatch):
    fake_refua = _install_fake_refua(monkeypatch)
    complex_spec, _, _, has_boltzgen = server._build_complex_from_spec(
        name="antibody",
        base_dir=None,
        entities=[
            {
                "type": "antibody",
                "ids": ["H", "L"],
                "heavy_cdr_lengths": [12, 10, 14],
                "light_cdr_lengths": "10,9,9",
                "heavy_spec": "{h1}G",
                "heavy_template_values": {"h1": "AA"},
            }
        ],
        boltz_mol_dir=None,
    )

    assert has_boltzgen is True
    assert len(fake_refua.BinderDesigns.antibody_calls) == 1
    call = fake_refua.BinderDesigns.antibody_calls[0]
    assert call["heavy_id"] == "H"
    assert call["light_id"] == "L"
    assert call["heavy_cdr_lengths"] == (12, 10, 14)
    assert call["light_cdr_lengths"] == (10, 9, 9)

    assert len(complex_spec.added) == 2
    heavy, light = complex_spec.added
    assert heavy.spec == "{h1}G"
    assert heavy.template_values == {"h1": "AA"}
    assert light.spec == "light-default"



from __future__ import annotations

import threading
import time

from mcp.types import CallToolResult

import refua_mcp.server as server


def test_recommend_poll_seconds_without_estimate_scales_by_queue() -> None:
    assert server._recommend_poll_seconds(None, queue_position=0) == 30
    assert server._recommend_poll_seconds(None, queue_position=1) == 45
    assert server._recommend_poll_seconds(None, queue_position=6) == 120


def test_recommend_poll_seconds_with_estimate_uses_floor_and_cap() -> None:
    assert server._recommend_poll_seconds(5, queue_position=0) == 30
    assert server._recommend_poll_seconds(60, queue_position=0) == 30
    assert server._recommend_poll_seconds(300, queue_position=0) == 105
    assert server._recommend_poll_seconds(1_200, queue_position=0) == 120


def test_task_support_mode_defaults() -> None:
    assert server._task_support_mode("refua_fold") == "optional"
    assert server._task_support_mode("refua_affinity") == "optional"
    assert server._task_support_mode("refua_antibody_design") == "optional"
    assert server._task_support_mode("refua_protein_properties") == "optional"
    assert server._task_support_mode("refua_job") == "optional"
    expected = "optional" if server._CLINICAL_AVAILABLE else "forbidden"
    assert server._task_support_mode("refua_clinical_simulator") == expected
    data_expected = "optional" if server._DATA_AVAILABLE else "forbidden"
    assert server._task_support_mode("refua_data_query") == data_expected
    preclinical_expected = "optional" if server._PRECLINICAL_AVAILABLE else "forbidden"
    assert server._task_support_mode("refua_preclinical_workup") == preclinical_expected
    assert (
        server._task_support_mode("refua_preclinical_release_assess")
        == preclinical_expected
    )
    assert server._task_support_mode("unknown_tool") == "forbidden"


def test_normalize_task_tool_result_structured_dict() -> None:
    result = server._normalize_task_tool_result({"ok": True})
    assert isinstance(result, CallToolResult)
    assert result.isError is False
    assert result.structuredContent == {"ok": True}


def test_long_poll_sleep_seconds_bounds() -> None:
    assert server._long_poll_sleep_seconds({"recommended_poll_seconds": 1}, 20) == 5
    assert server._long_poll_sleep_seconds({"recommended_poll_seconds": 300}, 20) == 20
    assert server._long_poll_sleep_seconds({"recommended_poll_seconds": 30}, 20) == 20


def test_poll_job_until_terminal_waits_once_then_returns(monkeypatch) -> None:
    snapshots = iter(
        [
            {"status": "running", "recommended_poll_seconds": 40},
            {"status": "success", "result": {"ok": True}},
        ]
    )
    now = {"t": 100.0}
    sleep_calls: list[float] = []

    def fake_snapshot(job_id: str, include_result: bool) -> dict[str, object]:
        return dict(next(snapshots))

    def fake_time() -> float:
        return now["t"]

    def fake_sleep(seconds: float) -> None:
        sleep_calls.append(seconds)
        now["t"] += seconds

    monkeypatch.setattr(server, "_job_snapshot", fake_snapshot)
    monkeypatch.setattr(server.time, "time", fake_time)
    monkeypatch.setattr(server.time, "sleep", fake_sleep)

    result = server._poll_job_until_terminal(
        "job-1",
        include_result=True,
        wait_for_terminal_seconds=120,
    )
    assert result["status"] == "success"
    assert sleep_calls == [40]


def test_poll_job_until_terminal_respects_timeout(monkeypatch) -> None:
    now = {"t": 100.0}
    sleep_calls: list[float] = []
    call_count = {"n": 0}

    def fake_snapshot(job_id: str, include_result: bool) -> dict[str, object]:
        call_count["n"] += 1
        return {"status": "running", "recommended_poll_seconds": 30}

    def fake_time() -> float:
        return now["t"]

    def fake_sleep(seconds: float) -> None:
        sleep_calls.append(seconds)
        now["t"] += seconds

    monkeypatch.setattr(server, "_job_snapshot", fake_snapshot)
    monkeypatch.setattr(server.time, "time", fake_time)
    monkeypatch.setattr(server.time, "sleep", fake_sleep)

    result = server._poll_job_until_terminal(
        "job-2",
        include_result=False,
        wait_for_terminal_seconds=10,
    )
    assert result["status"] == "running"
    assert sleep_calls == [10]
    assert call_count["n"] == 2


def test_refua_job_wait_for_terminal_uses_long_poll(monkeypatch) -> None:
    expected = {"status": "success", "result": {"done": True}}

    def fake_long_poll(
        job_id: str,
        *,
        include_result: bool,
        wait_for_terminal_seconds: float,
    ) -> dict[str, object]:
        assert job_id == "abc123"
        assert include_result is True
        assert wait_for_terminal_seconds == 9
        return expected

    monkeypatch.setattr(server, "_poll_job_until_terminal", fake_long_poll)
    result = server.refua_job(
        "abc123",
        include_result=True,
        wait_for_terminal_seconds=9,
    )
    assert result == expected


def _wait_for_terminal(job_id: str, *, timeout: float = 5.0) -> dict[str, object]:
    deadline = time.time() + timeout
    snapshot = server._job_snapshot(job_id, include_result=True)
    while snapshot["status"] in {"queued", "running"}:
        if time.time() >= deadline:
            raise TimeoutError(f"Job did not reach terminal state: {job_id}")
        time.sleep(0.01)
        snapshot = server._job_snapshot(job_id, include_result=True)
    return snapshot


def test_cancel_job_marks_queued_job_cancelled() -> None:
    gate = threading.Event()

    def blocking_runner() -> dict[str, bool]:
        gate.wait(timeout=2.0)
        return {"ok": True}

    first = server._submit_job("test_tool", blocking_runner, queue_timeout_seconds=30)
    second = server._submit_job(
        "test_tool",
        lambda: {"ok": True},
        queue_timeout_seconds=30,
    )

    snapshot = server._cancel_job(
        second,
        reason={
            "code": "job_cancelled",
            "message": "cancelled from test",
            "retryable": True,
        },
    )
    assert snapshot["status"] == "cancelled"
    assert snapshot["error"]["code"] == "job_cancelled"

    gate.set()
    _wait_for_terminal(first)


def test_queue_timeout_fails_stale_job() -> None:
    gate = threading.Event()

    def blocking_runner() -> dict[str, bool]:
        gate.wait(timeout=2.0)
        return {"ok": True}

    first = server._submit_job("test_tool", blocking_runner, queue_timeout_seconds=30)
    second = server._submit_job(
        "test_tool",
        lambda: {"ok": True},
        queue_timeout_seconds=0.01,
    )

    time.sleep(0.05)
    gate.set()

    first_snapshot = _wait_for_terminal(first)
    assert first_snapshot["status"] in {"success", "cancelled", "error"}

    second_snapshot = _wait_for_terminal(second)
    assert second_snapshot["status"] == "error"
    assert second_snapshot["error"]["code"] == "queue_timeout"


def test_build_task_runner_supports_protein_properties(monkeypatch) -> None:
    captured: dict[str, object] = {}

    def fake_tool(**kwargs):
        captured.update(kwargs)
        return {"ok": True}

    monkeypatch.setattr(server, "refua_protein_properties", fake_tool)
    runner = server._build_task_runner(
        "refua_protein_properties",
        {"sequence": "MKTAYIAK"},
    )
    assert runner is not None
    assert runner() == {"ok": True}
    assert captured["sequence"] == "MKTAYIAK"


def test_build_task_runner_supports_refua_job(monkeypatch) -> None:
    captured: dict[str, object] = {}

    def fake_tool(**kwargs):
        captured.update(kwargs)
        return {"status": "running"}

    monkeypatch.setattr(server, "refua_job", fake_tool)
    runner = server._build_task_runner(
        "refua_job",
        {"job_id": "job-123", "include_result": True},
    )
    assert runner is not None
    assert runner() == {"status": "running"}
    assert captured["job_id"] == "job-123"
    assert captured["include_result"] is True
